/*==================================================================*/
/* CS6490 Graphics Support Code                                     */
/* Implementation File                                              */
/* Version 2.0       						    */
/*==================================================================*/


#ifdef GTWINDOW
#include <GL/gl.h>
#include <GL/glu.h>
#include <aux.h>
#endif

#include <stdio.h>
#include<stdlib.h>
#include <malloc.h>

#include "gtGraphics.h"

typedef unsigned char Byte;

/*==================================================================*/
/* gtImage Data Type Implementation                                 */
/*------------------------------------------------------------------*/

long gtArray3DTo1D ( int chan, int w, int sizeW, int h, int sizeH )
{
   /* convert 3D array coordinates to a single lookup value */
   return ( chan * sizeW * sizeH + w * sizeH + h );
}

/*-----------------------------------*/

long gtImageValueCount ( gtImage *I )
{
   return ( I->w * I->h * 3 );
}

/*-----------------------------------*/

void gtImageInit ( gtImage *I )
{
   I->w = I->h = 0;
   I->values = (int *) 0;
}

/*-----------------------------------*/

void gtImageDestroy ( gtImage *I )
{
   free( I->values );
}

/*-----------------------------------*/

void gtImageGetSize ( gtImage *I, int *w, int *h )
{
   *w = I->w;
   *h = I->h;
}

/*-----------------------------------*/

void gtImageSetSize ( gtImage *I, int w, int h )
{
   long count;

   free( I->values );

   I->w = w;
   I->h = h;
   count = gtImageValueCount( I );
   I->values = (int *) calloc( (unsigned int) count, sizeof( int ) );
   if ( ! I->values ) {
      fprintf( stderr, "gtImage not allocated\n" );
      exit( 2 );
   }
}

/*-----------------------------------*/

FILE* gtImageWritePPM ( gtImage *I, char *filename )
{

   /*-------------------------------------------------------------*/
   /* Note:  This outputs the "raw" PPM format, meaning it is     */
   /*        binary and not very pretty for human eyes.  It       */
   /*        could be made into a human ascii version but it      */
   /*        would produce files four times bigger and, come on   */
   /*        now, can you *really* get much from looking at it?   */
   /*        This *is* a *graphics* library!  Use 'xv' already.   */
   /*-------------------------------------------------------------*/

   int row, col, chan;
   int r, g, b;
   Byte br, bg, bb;

   FILE *fp = fopen( filename, "w" );
   if ( ! fp ) return fp;

   /* output header */
   fprintf( fp, "P6\n" );
   fprintf( fp, "%d %d\n", I->w, I->h );
   fprintf( fp, "%d\n", 255 );

   /* write image values (only handles 8 bits) */
   for ( row = 0; row < I->h; row++ )
      for ( col = 0; col < I->w; col++ ) {
         gtImageReadPixel( I, col, I->h - row - 1, &r, &g, &b );
	 br = (Byte) r;
	 bg = (Byte) g;
	 bb = (Byte) b;
         if ( fwrite( &br, sizeof( Byte ), 1, fp ) != sizeof( Byte ) )
            return (FILE*) NULL;
         if ( fwrite( &bg, sizeof( Byte ), 1, fp ) != sizeof( Byte ) )
            return (FILE*) NULL;
         if ( fwrite( &bb, sizeof( Byte ), 1, fp ) != sizeof( Byte ) )
            return (FILE*) NULL;
      }

   fclose( fp );
   return fp;
}

/*-----------------------------------*/

void gtImageReadPixel ( gtImage *I, int w, int h, int *r, int *g, int *b )
{
   long index = gtArray3DTo1D( GT_RED, w, I->w, h, I->h );
   *r = I->values[ index ];

   index = gtArray3DTo1D( GT_GREEN, w, I->w, h, I->h );
   *g = I->values[ index ];

   index = gtArray3DTo1D( GT_BLUE, w, I->w, h, I->h );
   *b = I->values[ index ];
}

/*-----------------------------------*/

void gtImageWritePixel ( gtImage *I, int w, int h, int r, int g, int b, float opacity)
{
   long index;
   index = gtArray3DTo1D( GT_RED, w, I->w, h, I->h );
   I->values[ index ] = I->values[index]+opacity*r;
if (I->values[ index ]>255) I->values[ index ]=255;

   index = gtArray3DTo1D( GT_GREEN, w, I->w, h, I->h );
   I->values[ index ] = I->values[index]+opacity*g;
if (I->values[ index ]>255) I->values[ index ]=255;

   index = gtArray3DTo1D( GT_BLUE, w, I->w, h, I->h );
   I->values[ index ] = I->values[index]+opacity*b;
if (I->values[ index ]>255) I->values[ index ]=255;
}

/*==================================================================*/
/* Single Window Interface                                          */
/*------------------------------------------------------------------*/

static int /* GTWriteMode */ GTCurrentMode;
static gtImage GTCurrentImage;

void gtBeginGraphics ( int w, int h )
{

#ifdef GTWINDOW

   /* TO DO:  needs to allow placement of the window */

   /* Make a 3D OpenGL window to be used for 2D   */
   /* as well as set up the single window system. */

   /* set up a window with RGB color and one buffer */
   auxInitDisplayMode( AUX_RGB | AUX_SINGLE );

   /* set it's screen location */
   auxInitPosition( 0, 0, w, h );

   /* bring up the window with a name */
   auxInitWindow( "gtOut" );

   /* make it fake a 2D screen */
   gluOrtho2D( -0.5, w - 0.5, -0.5, h - 0.5 );

#endif

   /* set the single window state */
   GTCurrentMode = GT_OVERWRITE;

   /* build a frame buffer for reading and writing */
   gtImageInit( &GTCurrentImage );
   gtImageSetSize( &GTCurrentImage, w, h );

   gtClear( 0, 0, 0 );

}

/*-----------------------------------*/

void gtGetFramebufferSize ( int *w, int *h )
{
   *w = GTCurrentImage.w;
   *h = GTCurrentImage.h;
}

/*-----------------------------------*/

void gtEndGraphics ()
{

#ifdef GTWINDOW
   /* TO DO:  needs to remove the window */
#endif

   gtImageDestroy( &GTCurrentImage );
}

/*-----------------------------------*/

void gtClear ( int r, int g, int b )
{
   int i, j;

#ifdef GTWINDOW
   glClearColor( (float) r, (float) g, (float) b, 0.0 );
   glClear( GL_COLOR_BUFFER_BIT );
#endif

   /* clear image */
   for ( i = 0; i < GTCurrentImage.w; i++ )
      for ( j = 0; j < GTCurrentImage.h; j++ )
         gtImageWritePixel( &GTCurrentImage, i, j, r, g, b, 1 );
}

/*-----------------------------------*/

void gtFlush ()
{

#ifdef GTWINDOW
   glFlush ();
#endif

}

/*-----------------------------------*/

void gtSetWriteMode ( int /* GTWriteMode */ mode )
{
   GTCurrentMode = mode;
}

/*-----------------------------------*/

int gtXOR ( int a, int b )
{
   return ( a ^ b);
}

/*-----------------------------------*/

void gtWriteFramebuffer ( char *filename ) {

   FILE *fp;

   fp = gtImageWritePPM ( &GTCurrentImage, filename );
   if ( ! fp ) {
      fprintf( stderr, "Frame Buffer Write Error!  Buffer not written.\n" );
   }
}

/*-----------------------------------*/

void gtWritePixel ( int x, int y, int r, int g, int b, float opacity )
{
   int or, og, ob;

   if ( x < 0 || x >= GTCurrentImage.w || y < 0 || y >= GTCurrentImage.h )
   {
     fprintf( stderr, "Attempted to write a pixel outside the image " );
     fprintf( stderr, "bounds: x = %d, y = %d\n", x, y );
     return;
   }

   if ( GTCurrentMode == GT_XOR )
   {
      gtImageReadPixel( &GTCurrentImage, x, y, &or, &og, &ob );
      r = gtXOR ( r, or );
      g = gtXOR ( g, og );
      b = gtXOR ( b, ob );
   }

#ifdef GTWINDOW
   glColor3f( r / 255.0, g / 255.0, b / 255.0 );
   glBegin( GL_POINTS );
     glVertex2i( x, y );
   glEnd();
#endif

   gtImageWritePixel( &GTCurrentImage, x, y, r, g, b ,opacity);
}

/*-----------------------------------*/

void gtRedraw ( void )
{

#ifdef GTWINDOW

   int x, y;
   int r, g, b;

   /* Reapply fake 2D screen - problems occur otherwise */
   /* Nothing is redrawn if this is used */
   /*
   gluOrtho2D( -0.5, GTCurrentImage.w - 0.5, -0.5, GTCurrentImage.h - 0.5 );
   */

   glBegin( GL_POINTS );

      for ( x = 0; x < GTCurrentImage.w; x++ ) {
         for ( y = 0; y < GTCurrentImage.h; y++ ) {
            gtImageReadPixel( &GTCurrentImage, x, y, &r, &g, &b );
            glColor3f( r / 255.0, g / 255.0, b / 255.0 );
            glVertex2i( x, y );
	 }
      }

   glEnd();

   glFlush();

#endif

}

/*-----------------------------------*/

void gtWaitOnEscape ()
{

#ifdef GTWINDOW

   /* This function is having trouble */
   /* auxMainLoop( gtRedraw ); */

   auxMainLoop( 0 );

#endif

}

/*-----------------------------------*/

